﻿using Eneter.Messaging.DataProcessing.Serializing;
using Eneter.Messaging.Diagnostic;
using Eneter.Messaging.EndPoints.Rpc;
using Eneter.Messaging.MessagingSystems.Composites.AuthenticatedConnection;
using Eneter.Messaging.MessagingSystems.ConnectionProtocols;
using Eneter.Messaging.MessagingSystems.MessagingSystemBase;
using Eneter.Messaging.MessagingSystems.TcpMessagingSystem;
using Eneter.SecureRemotePassword;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Security.Cryptography;
using System.Threading;

namespace Service
{
    // Interface declaring the RPC service.
    public interface ICalculator
    {
        double Sum(double a, double b);
    }

    // Implementation of the RPC service.
    internal class Calculator : ICalculator
    {
        public double Sum(double a, double b)
        {
            double aResult = a + b;

            // Example how to get a user name from within the method.
            // Note: it must be called from the same thread as this method was originally called.
            string aUserName = SrpAuthentication.GetUserName();

            // Didplay results.
            Console.WriteLine("{0}: {1} + {2} = {3}", aUserName, a, b, aResult);

            return aResult;
        }
    }

    // Simulates database of users.
    internal static class UserDb
    {
        // Reperesents user authentication data stored in DB.
        public class User
        {
            public User(string userName, byte[] salt, byte[] verifier)
            {
                UserName = userName;
                Salt = salt;
                Verifier = verifier;
            }

            public string UserName { get; private set; }
            public byte[] Salt { get; private set; }
            public byte[] Verifier { get; private set; }
        }

        private static HashSet<User> myUsers = new HashSet<User>();

        public static void CreateUser(string userName, string password)
        {
            // Generate the random salt.
            byte[] s = SRP.s();

            // Compute private key from password nad salt.
            byte[] x = SRP.x(password, s);

            // Compute verifier.
            byte[] v = SRP.v(x);

            // Store user name, salt and the verifier.
            // Note: do not store password nor the private key in database!
            User aUser = new User(userName, s, v);
            lock (myUsers)
            {
                myUsers.Add(aUser);
            }
        }

        public static User GetUser(string userName)
        {
            lock (myUsers)
            {
                User aUser = myUsers.FirstOrDefault(x => x.UserName == userName);
                return aUser;
            }
        }
    }

    // Encapsulates the SRP authentication.
    internal static class SrpAuthentication
    {
        // Connection context for each connected client.
        private class ConnectionContext
        {
            public ConnectionContext(string responseReceiverId, string userName)
            {
                ResponseReceiverId = responseReceiverId;
                UserName = userName;
                ThreadId = -1;
            }

            // Identifies the connection session.
            public string ResponseReceiverId { get; private set; }

            // Identifies the thread which is right now handling the request.
            public int ThreadId { get; set; }

            // Login name.
            public string UserName { get; private set; }

            // SRP values used during the authentication process.
            public byte[] K { get; set; }
            public byte[] A { get; set; }
            public byte[] B { get; set; }
            public byte[] s { get; set; }

            // Serializer to serialize/deserialize messages once
            // the authenticated connection is established.
            // It uses the session key (calculated during RSP authentication) to encrypt/decrypt messages. 
            public ISerializer Serializer { get; set; }
        }

        // List of connected clients.
        private static List<ConnectionContext> myConnections = new List<ConnectionContext>();

        // Returns SRP response for login request.
        public static object GetLoginResponseMessage(
            string responseReceiverId, object loginRequestMessage)
        {
            // Deserialize the login request.
            ISerializer aSerializer = new BinarySerializer();
            LoginRequestMessage aLoginRequest = aSerializer.Deserialize<LoginRequestMessage>(loginRequestMessage);

            // Try to find the user in database.
            UserDb.User aUser = UserDb.GetUser(aLoginRequest.UserName);
            if (aUser != null &&
                SRP.IsValid_A(aLoginRequest.A))
            {
                // Generate random service private ephemeral value.
                byte[] b = SRP.b();

                // Calculate service public ephemeral value.
                byte[] B = SRP.B(b, aUser.Verifier);

                // Calculate random scrambling value.
                byte[] u = SRP.u(aLoginRequest.A, B);

                // Calculate session key.
                byte[] K = SRP.K_Service(aLoginRequest.A, aUser.Verifier, u, b);

                // Prepare response message for the client.
                // Note: client is then supposed to calculate the session key
                //       and send back the message proving it was able to calculate the same session key.
                LoginResponseMessage aLoginResponse = new LoginResponseMessage();
                aLoginResponse.s = aUser.Salt; // user salt
                aLoginResponse.B = B;       // service public ephemeral value
                object aLoginResponseMessage = aSerializer.Serialize<LoginResponseMessage>(aLoginResponse);

                // Store the connection context.
                ConnectionContext aConnection = new ConnectionContext(responseReceiverId, aUser.UserName);
                aConnection.A = aLoginRequest.A;
                aConnection.B = B;
                aConnection.K = K;
                aConnection.s = aUser.Salt;
                lock (myConnections)
                {
                    myConnections.Add(aConnection);
                }

                // Send the response to the client.
                return aLoginResponseMessage;
            }

            // The client will be disconnected.
            return null;
        }

        // Checks if the connecting client was able to calculate the same
        // session key as the service.
        // If the user entered an incorrect password the calculation
        // will be different.
        public static bool Authenticate(string responseReceiverId, object M1)
        {
            ConnectionContext aConnection;
            lock (myConnections)
            {
                aConnection = myConnections.FirstOrDefault(
                    x => x.ResponseReceiverId == responseReceiverId);
            }
            if (aConnection != null)
            {
                // Proving message from the client.
                byte[] aClientM1 = (byte[])M1;

                // Service calculates the proving message too.
                byte[] aServiceM1 = SRP.M1(aConnection.A, aConnection.B, aConnection.K);

                // If both messages are equql then it means the client proved its identity
                // and the connection can be established.
                if (aServiceM1.SequenceEqual(aClientM1))
                {
                    // Create serializer.
                    Rfc2898DeriveBytes anRfc = new Rfc2898DeriveBytes(aConnection.K, aConnection.s, 1000);
                    ISerializer aSerializer = new AesSerializer(new BinarySerializer(), anRfc, 256);

                    // Store serializer which will encrypt using the calculated key.
                    aConnection.Serializer = aSerializer;

                    // Clean properties which are not needed anymore.
                    aConnection.A = null;
                    aConnection.B = null;
                    aConnection.K = null;
                    aConnection.s = null;

                    return true;
                }

                lock (myConnections)
                {
                    myConnections.RemoveAll(
                        x => x.ResponseReceiverId == responseReceiverId);
                }
            }

            return false;
        }

        // Cancel the authentication sequense.
        public static void CancelAuthentication(string responseReceiverId)
        {
            lock (myConnections)
            {
                myConnections.RemoveAll(x => x.ResponseReceiverId == responseReceiverId);
            }
        }

        // Removes the connection.
        public static void CloseConnection(string responseReceiverId)
        {
            lock (myConnections)
            {
                ConnectionContext aConnection = myConnections.FirstOrDefault(
                    x => x.ResponseReceiverId == responseReceiverId);
                if (aConnection != null)
                {
                    myConnections.Remove(aConnection);
                }
            }
        }

        // Returns user name for the specific response receiver id.
        public static string GetUserName(string responseReceiverId)
        {
            string aUserName = "";
            lock (myConnections)
            {
                ConnectionContext aConnection = myConnections.FirstOrDefault(
                    x => x.ResponseReceiverId == responseReceiverId);
                if (aConnection != null)
                {
                    aUserName = aConnection.UserName;
                }
            }

            return aUserName;
        }

        // Returns user name for the current thread.
        public static string GetUserName()
        {
            string aUserName = "";
            lock (myConnections)
            {
                ConnectionContext aConnection = myConnections.FirstOrDefault(
                    x => x.ThreadId == Thread.CurrentThread.ManagedThreadId);
                if (aConnection != null)
                {
                    aUserName = aConnection.UserName;
                }
            }

            return aUserName;
        }

        // This method is called when the RPC service needs to serialize or deserialize
        // request/response for the particular user and connection.
        public static ISerializer GetConnectionSpecificSerializer(string responseReceiverId)
        {
            ConnectionContext aUserContext;
            lock (myConnections)
            {
                aUserContext = myConnections.FirstOrDefault(
                    x => x.ResponseReceiverId == responseReceiverId);
            }
            if (aUserContext != null)
            {
                // Store also the thread id so that it is possible to identify a user
                // from within the RPC method.
                // Note: this is a workaround which will be resolved in a future Eneter release.
                aUserContext.ThreadId = Thread.CurrentThread.ManagedThreadId;

                // Return the connection specific serializer.
                return aUserContext.Serializer;
            }

            throw new InvalidOperationException("Failed to get serializer for the given connection.");
        }
    }

    class Program
    {
         static void Main(string[] args)
        {
            // Display Eneter errors and warnings on the console.
            //EneterTrace.TraceLog = Console.Out;

            // Simulate database of users.
            UserDb.CreateUser("Peter", "pwd123");
            UserDb.CreateUser("Frank", "pwd456");

            try
            {
                // Instantiate class which provides the service.
                ICalculator aCalculator = new Calculator();

                // Create RPC service.
                IRpcFactory aFactory = new RpcFactory()
                {
                    // Note: this allows to encrypt/decrypt messages for each client
                    // individualy based on calculated session key.
                    SerializerProvider = OnGetSerializer
                };
                IRpcService<ICalculator> aRpcService =
                    aFactory.CreateSingleInstanceService<ICalculator>(aCalculator);

                // Use TCP for the communication.
                IMessagingSystemFactory aUnderlyingMessaging =
                    new TcpMessagingSystemFactory(new EasyProtocolFormatter());
                
                // Use communication which is authenticated via SRP.
                IMessagingSystemFactory aMessaging =
                    new AuthenticatedMessagingFactory(aUnderlyingMessaging,
                    OnGetLoginResponseMessage,
                    OnAuthenticate,
                    OnAuthenticationCancelled);

                // Crete input channel and attach it to the receiver to start listening.
                IDuplexInputChannel aInputChannel =
                    aMessaging.CreateDuplexInputChannel("tcp://127.0.0.1:8033/");
                aInputChannel.ResponseReceiverConnected += OnClientConnected;
                aInputChannel.ResponseReceiverDisconnected += OnClientDisconnected;

                // Attach inout channel to the RPC service and start listning.
                aRpcService.AttachDuplexInputChannel(aInputChannel);

                Console.WriteLine("RPC service is running. Press ENTER to stop.");
                Console.ReadLine();

                // Detach input channel to stop the listening thread.
                aRpcService.DetachDuplexInputChannel();
            }
            catch (Exception err)
            {
                EneterTrace.Error("RPC Service failed.", err);
            }
        }

        private static object OnGetLoginResponseMessage(string channelId,
                string responseReceiverId, object loginRequestMessage)
        {
            return SrpAuthentication.GetLoginResponseMessage(responseReceiverId, loginRequestMessage);
        }

        private static bool OnAuthenticate(string channelId, string responseReceiverId,
                object login, object handshakeMessage, object M1)
        {
            return SrpAuthentication.Authenticate(responseReceiverId, M1);
        }

        private static void OnAuthenticationCancelled(
                string channelId, string responseReceiverId, object loginMessage)
        {
            SrpAuthentication.CancelAuthentication(responseReceiverId);
        }


        private static void OnClientConnected(object sender, ResponseReceiverEventArgs e)
        {
            string aUserName = SrpAuthentication.GetUserName(e.ResponseReceiverId);
            Console.WriteLine(aUserName + " is logged in.");
        }

        // Remove the connection context if the client disconnects once the connection
        // was established after the successful authentication.
        private static void OnClientDisconnected(object sender, ResponseReceiverEventArgs e)
        {
            string aUserName = SrpAuthentication.GetUserName(e.ResponseReceiverId);
            SrpAuthentication.CloseConnection(e.ResponseReceiverId);
            Console.WriteLine(aUserName + " is logged out.");
        }

        // It is called by MultiTypedReceiver whenever it sends or receive a message from a connected client.
        // It returns the serializer for the particular connection (which uses the agreed session key).
        private static ISerializer OnGetSerializer(string responseReceiverId)
        {
            return SrpAuthentication.GetConnectionSpecificSerializer(responseReceiverId);
        }
       
    }
}
