﻿using Eneter.Messaging.DataProcessing.Serializing;
using Eneter.Messaging.Diagnostic;
using Eneter.Messaging.EndPoints.Rpc;
using Eneter.Messaging.MessagingSystems.Composites.AuthenticatedConnection;
using Eneter.Messaging.MessagingSystems.ConnectionProtocols;
using Eneter.Messaging.MessagingSystems.MessagingSystemBase;
using Eneter.Messaging.MessagingSystems.TcpMessagingSystem;
using Eneter.Messaging.Threading.Dispatching;
using Eneter.SecureRemotePassword;
using System;
using System.IO;
using System.Security.Cryptography;
using System.Windows.Forms;

namespace WindowsFormClient
{
    public partial class Form1 : Form
    {
        // Interface declaring the RPC service.
        public interface ICalculator
        {
            double Sum(double a, double b);
        }

        // Encapsulates the SRP authentication for client.
        private class SrpClientAuthentication
        {
            private LoginRequestMessage myLoginRequest;
            private byte[] myPrivateKey_a;
            private ISerializer mySerializer;

            // It is called by the AuthenticationMessaging to get the login message.
            public object GetLoginRequestMessage(string userName)
            {
                myPrivateKey_a = SRP.a();
                byte[] A = SRP.A(myPrivateKey_a);

                myLoginRequest = new LoginRequestMessage();
                myLoginRequest.UserName = userName;
                myLoginRequest.A = A;

                // Serializer to serialize LoginRequestMessage.
                ISerializer aSerializer = new BinarySerializer();
                object aSerializedLoginRequest = aSerializer.Serialize<LoginRequestMessage>(myLoginRequest);

                // Send the login request to start negotiation about the session key.
                return aSerializedLoginRequest;
            }

            // It is called by the AuthenticationMessaging to handle the LoginResponseMessage received
            // from the service.
            public object GetProveMessage(
                string userPassword, object loginResponseMessage)
            {
                // Deserialize LoginResponseMessage.
                ISerializer aSerializer = new BinarySerializer();
                LoginResponseMessage aLoginResponse =
                    aSerializer.Deserialize<LoginResponseMessage>(loginResponseMessage);

                // Calculate scrambling parameter.
                byte[] u = SRP.u(myLoginRequest.A, aLoginResponse.B);

                if (SRP.IsValid_B_u(aLoginResponse.B, u))
                {
                    // Calculate user private key.
                    byte[] x = SRP.x(userPassword, aLoginResponse.s);

                    // Calculate the session key which will be used for the encryption.
                    // Note: if everything is then this key will be the same as on the service side.
                    byte[] K = SRP.K_Client(aLoginResponse.B, x, u, myPrivateKey_a);

                    // Create serializer for encrypting the communication.
                    Rfc2898DeriveBytes anRfc = new Rfc2898DeriveBytes(K, aLoginResponse.s, 1000);
                    mySerializer = new AesSerializer(new BinarySerializer(), anRfc, 256);

                    // Create M1 message to prove that the client has the correct session key.
                    byte[] M1 = SRP.M1(myLoginRequest.A, aLoginResponse.B, K);
                    return M1;
                }

                // Close the connection with the service.
                return null;
            }

            // It is called whenever the client sends or receives the message from the service.
            // It will return the serializer which serializes/deserializes messages using
            // the connection password.
            public ISerializer GetSerializer()
            {
                return mySerializer;
            }
        }

        private SrpClientAuthentication mySrpAuthentication = new SrpClientAuthentication();
        private IRpcClient<ICalculator> myRpcClient;
        
        

        public Form1()
        {
            InitializeComponent();
            EnableUiControls(false);
        }

        private void Form1_FormClosed(object sender, FormClosedEventArgs e)
        {
            CloseConnection();
        }

        private void OpenConnection()
        {
            IMessagingSystemFactory anUnderlyingMessaging = new TcpMessagingSystemFactory(new EasyProtocolFormatter());
            IMessagingSystemFactory aMessaging =
                new AuthenticatedMessagingFactory(anUnderlyingMessaging, OnGetLoginRequestMessage, OnGetProveMessage)
            {
                // Timeout for the authentication.
                // If the value is -1 then it is infinite (e.g. for debugging purposses)
                AuthenticationTimeout = TimeSpan.FromMilliseconds(30000)
            };

            IDuplexOutputChannel anOutputChannel = aMessaging.CreateDuplexOutputChannel("tcp://127.0.0.1:8033/");
            anOutputChannel.ConnectionClosed += OnConnectionClosed;

            IRpcFactory aFactory = new RpcFactory()
            {
                SerializerProvider = OnGetSerializer
            };
            myRpcClient = aFactory.CreateClient<ICalculator>();

            try
            {
                // Attach output channel and be able to call RPC.
                myRpcClient.AttachDuplexOutputChannel(anOutputChannel);
                EnableUiControls(true);
            }
            catch
            {
                MessageBox.Show("Incorrect user name or password.",
                    "Login Failure", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        // It is called if the service closes the connection.
        private void OnConnectionClosed(object sender, DuplexChannelEventArgs e)
        {
            // It does not have to be invoked in the main UI thread.
            // So in order to manipulate UI controls dispatch the call
            // to the main UI thread.
            if (InvokeRequired)
            {
                Action aCloseConnection = CloseConnection;
                Invoke(aCloseConnection, null);
            }
            else
            {
                CloseConnection();
            }
        }

        private void CloseConnection()
        {
            // Detach output channel and release the thread listening to responses.
            if (myRpcClient != null && myRpcClient.IsDuplexOutputChannelAttached)
            {
                myRpcClient.DetachDuplexOutputChannel();
            }
            EnableUiControls(false);
        }

        private void EnableUiControls(bool isLoggedIn)
        {
            LoginTextBox.Enabled = !isLoggedIn;
            PasswordTextBox.Enabled = !isLoggedIn;
            LoginBtn.Enabled = !isLoggedIn;

            LogoutBtn.Enabled = isLoggedIn;
            Number1TextBox.Enabled = isLoggedIn;
            Number2TextBox.Enabled = isLoggedIn;
            CalculateBtn.Enabled = isLoggedIn;
            ResultTextBox.Enabled = isLoggedIn;
        }

        // It is called by the AuthenticationMessaging to get the login message.
        private object OnGetLoginRequestMessage(string channelId, string responseReceiverId)
        {
            string aUserName = LoginTextBox.Text;
            return mySrpAuthentication.GetLoginRequestMessage(aUserName);
        }

        // It is called by the AuthenticationMessaging to handle the LoginResponseMessage received
        // from the service.
        private object OnGetProveMessage(
            string channelId, string responseReceiverId, object loginResponseMessage)
        {
            string aUserPassword = PasswordTextBox.Text;
            return mySrpAuthentication.GetProveMessage(aUserPassword, loginResponseMessage);
        }

        // It is called whenever the client sends or receives the message from the service.
        // It will return the serializer which serializes/deserializes messages using
        // the connection password.
        private ISerializer OnGetSerializer(string responseReceiverId)
        {
            return mySrpAuthentication.GetSerializer();
        }


        private void CalculateBtn_Click(object sender, EventArgs e)
        {
            // Create message.
            double a = double.Parse(Number1TextBox.Text);
            double b = double.Parse(Number2TextBox.Text);
            double aResult = myRpcClient.Proxy.Sum(a, b);

            ResultTextBox.Text = aResult.ToString();
        }

        private void LoginBtn_Click(object sender, EventArgs e)
        {
            OpenConnection();
        }

        private void LogoutBtn_Click(object sender, EventArgs e)
        {
            CloseConnection();
        }
    }
}
